import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';

import 'resolver.dart';
import 'extensions.dart';

typedef Predicate<T> = bool Function(T a, T b);
final Predicate _neq = (a, b) => a != b;

class QueryAspect<T> {
  final Query<T> query;
  final Resolver resolver;
  T? _cachedResult;

  QueryAspect(this.query, this.resolver);

  bool shouldNotfiy(ValueNotifier stateNotifier) {
    final newValue = resolver.resolve(stateNotifier, query.params);
    final didChange = query.updateCriteria(_cachedResult, newValue);
    _cachedResult = newValue;

    return didChange;
  }
}

class Query<T> {
  final Object? params;
  final Predicate updateCriteria;

  Query([this.params, Predicate? updateCriteria])
      : this.updateCriteria = updateCriteria ?? _neq;

  T? exec(BuildContext context) {
    final r = context.getResolverForQueryType<T>();
    final container = r.getContainer(context);
    return r.resolve(container.stateNotifier, params);
  }

  T? subscribe(BuildContext context) {
    final r = context.getResolverForQueryType<T>();
    final container = r.getContainer(context);
    final element = context.getContainerHolder(container);

    final result = r.resolve(container.stateNotifier, params);
    final aspect = QueryAspect(this, r);
    aspect._cachedResult = result;

    context.dependOnInheritedElement(element, aspect: aspect);

    return result;
  }

  unsubscribe(Element element) {
    final r = element.getResolverForQueryType<T>();
    final container = r.getContainer(element);
    final el = element.getContainerHolder(container);
    el.unsubscribe(element);
  }
}

class QueryResolver<T, K> extends Resolver<T, K, Object> {
  Type get keyType => K;

  QueryResolver(
      ResolverController<T> ctrl, ResolverCallback<K, T, Object> callback)
      : super(ctrl, callback);
}
